from abc import ABC, abstractmethod
from logging import getLogger
from typing import Callable, Iterable, TYPE_CHECKING, List, Tuple

from more_itertools import zip_equal

from ..config import CONFIG as _CONFIG
from ..entity import Entity
from .result import SearchResultList
from .annotation import Annotations, Annotation, Correctness, Completeness
from .result import TaggedSentence, strip_prefix, strip_suffix

if TYPE_CHECKING:  # avoid circular dependency
    from ..document import WindowPosition
    from ..index import Result
    from ..entity import NerTag


LOGGER = getLogger(__name__)


RelevanceCriterion = Callable[[Annotation], bool]

Precision = float
Recall = float
FScore = float


class EvaluationResult(ABC):
    CONFIG = _CONFIG['search.EvaluationResult']
    BETA = CONFIG.getfloat('f_score_beta')

    @property
    @abstractmethod
    def precision(self) -> Precision:
        pass

    @property
    @abstractmethod
    def recall(self) -> Recall:
        pass

    @property
    def f_score(self) -> FScore:
        beta_squared = self.BETA ** 2
        numerator = (1 + beta_squared) * self.precision * self.recall
        denominator = (beta_squared * self.precision) + self.recall
        f_score = numerator / denominator if denominator > 0.0 else 0.0
        return f_score


class PartialNerEvaluationResult(EvaluationResult):
    def __init__(self, true_positives: int, false_positives: int, false_negatives: int):
        self.true_positives = true_positives
        self.false_positives = false_positives
        self.false_negatives = false_negatives

    @property
    def precision(self) -> Precision:
        if self.false_positives == 0:
            return 1.0
        return self.true_positives / (self.true_positives + self.false_positives)

    @property
    def recall(self) -> Recall:
        if self.false_negatives == 0:
            return 1.0
        return self.true_positives / (self.true_positives + self.false_negatives)


class NerEvaluationResult(EvaluationResult):
    CONFIG = _CONFIG['search.NerEvaluationResult']
    PREFERRED_TYPE = CONFIG['preferred_evaluation_type']

    def __init__(self, strict_result: PartialNerEvaluationResult,
                 fuzzy_result: PartialNerEvaluationResult):
        self.strict_result = strict_result
        self.fuzzy_result = fuzzy_result

    @property
    def preferred_result(self) -> PartialNerEvaluationResult:
        if self.PREFERRED_TYPE == 'strict':
            return self.strict_result
        elif self.PREFERRED_TYPE == 'fuzzy':
            return self.fuzzy_result
        else:
            raise ValueError(f'Unknown preferred evaluation type: {self.PREFERRED_TYPE}')

    @property
    def precision(self) -> Precision:
        return self.preferred_result.precision

    @property
    def recall(self) -> Recall:
        return self.preferred_result.recall

    @classmethod
    def from_tagged_sentences(
            cls, ground_truth: Iterable[TaggedSentence],
            predictions: Iterable[TaggedSentence]) -> 'NerEvaluationResult':

        def extract_named_entity_positions(
                tagged_sentence: TaggedSentence) -> List[Tuple['WindowPosition', 'NerTag']]:
            named_entity_positions: List[Tuple['WindowPosition', 'NerTag']] = []
            current_named_entity = None
            for position, ner_tag in enumerate((*tagged_sentence.ner_tags_tuple, 'O')):
                if current_named_entity is not None and strip_suffix(ner_tag) != 'I-':
                    window_position = (current_named_entity[0], position - 1)
                    named_entity_positions.append((window_position, current_named_entity[1]))
                if strip_suffix(ner_tag) == 'B-':
                    current_named_entity = (position, strip_prefix(ner_tag))
                if ner_tag == 'O':
                    current_named_entity = None
            return named_entity_positions

        def overlaps(first: 'WindowPosition', second: 'WindowPosition') -> bool:
            first, second = sorted([first, second])
            first_start, first_stop = first
            second_start, second_stop = second
            overlaps = first_start <= second_stop and second_start <= first_stop
            return overlaps

        strict_tp, strict_fp, strict_fn = 0, 0, 0
        fuzzy_tp, fuzzy_fp, fuzzy_fn = 0, 0, 0
        for ground_truth_tagged_sentence, prediction_tagged_sentence in zip_equal(
                ground_truth, predictions):
            ground_truth_positions = extract_named_entity_positions(ground_truth_tagged_sentence)
            prediction_positions = extract_named_entity_positions(prediction_tagged_sentence)

            for ground_truth_position_and_type in ground_truth_positions:
                ground_truth_position, ground_truth_type = ground_truth_position_and_type
                for prediction_position_and_type in prediction_positions:
                    if ground_truth_position_and_type == prediction_position_and_type:
                        strict_tp += 1
                        break
                else:
                    strict_fn += 1
                for prediction_position_and_type in prediction_positions:
                    prediction_position, prediction_type = prediction_position_and_type
                    if ground_truth_type == prediction_type:
                        if overlaps(ground_truth_position, prediction_position):
                            fuzzy_tp += 1
                            break
                else:
                    fuzzy_fn += 1

            for prediction_position_and_type in prediction_positions:
                prediction_position, prediction_type = prediction_position_and_type
                for ground_truth_position_and_type in ground_truth_positions:
                    if ground_truth_position_and_type == prediction_position_and_type:
                        break
                else:
                    strict_fp += 1
                for ground_truth_position_and_type in ground_truth_positions:
                    ground_truth_position, ground_truth_type = ground_truth_position_and_type
                    if ground_truth_type == prediction_type:
                        if overlaps(ground_truth_position, prediction_position):
                            break
                else:
                    fuzzy_fp += 1

        strict_result = PartialNerEvaluationResult(strict_tp, strict_fp, strict_fn)
        fuzzy_result = PartialNerEvaluationResult(fuzzy_tp, fuzzy_fp, fuzzy_fn)
        result = cls(strict_result, fuzzy_result)

        return result

    def __str__(self) -> str:
        summary = (
            'The results achieved Precision of '
            f'{100.0 * self.strict_result.precision:.2f}--{100.0 * self.fuzzy_result.precision:.2f}%'
            ', Recall of '
            f'{100.0 * self.strict_result.recall:.2f}--{100.0 * self.fuzzy_result.recall:.2f}%'
            ', and F-score of '
            f'{100.0 * self.strict_result.f_score:.2f}--{100.0 * self.fuzzy_result.f_score:.2f}%.'
        )
        return summary

    def __repr__(self) -> str:
        return str(self)


def default_relevance_criterion(annotation: Annotation) -> bool:
    if annotation.correctness == Correctness.FULL and annotation.completeness == Completeness.EXACT_OVERLAP:
        return True  # perfect match
    elif annotation.correctness == Correctness.PARTIAL and annotation.completeness == Completeness.EXACT_OVERLAP:
        return True  # different entity
    elif annotation.note == 'navíc pouze interpunknce':
        return True  # punctuation around entity
    return False


class IrEvaluationResult(EvaluationResult):
    CONFIG = _CONFIG['search.IrEvaluationResult']
    TOPN = CONFIG.getint('number_of_results')

    def __init__(self, result_list: SearchResultList, annotations: Annotations,
                 entities: Iterable[Entity],
                 relevance_criterion: RelevanceCriterion = default_relevance_criterion):
        self.result_list = result_list
        self.annotations = annotations.annotations
        self.entities = list(entities)
        self.relevance_criterion = relevance_criterion

    @property
    def number_of_total_relevant_results(self) -> int:
        number_of_total_relevant_results = 0
        for entity in self.entities:
            for annotation in self.annotations[entity].values():
                if self.relevance_criterion(annotation):
                    number_of_total_relevant_results += 1
        return number_of_total_relevant_results

    @property
    def number_of_found_relevant_results(self) -> int:
        number_of_found_relevant_results = 0
        for entity in self.entities:
            results = self._get_topn_results(entity)
            for (result_head, _) in results:
                if result_head not in self.annotations[entity]:
                    continue
                annotation = self.annotations[entity][result_head]
                if self.relevance_criterion(annotation):
                    number_of_found_relevant_results += 1
        return number_of_found_relevant_results

    @property
    def number_of_found_unannotated_results(self) -> int:
        number_of_found_unannotated_results = 0
        for entity in self.entities:
            results = self._get_topn_results(entity)
            for (result_head, _) in results:
                if result_head not in self.annotations[entity]:
                    number_of_found_unannotated_results += 1
        return number_of_found_unannotated_results

    @property
    def number_of_found_results(self) -> int:
        number_of_found_results = 0
        for entity in self.entities:
            results = self._get_topn_results(entity)
            for (result_head, _) in results:
                number_of_found_results += 1
        return number_of_found_results

    def _get_topn_results(self, entity: Entity) -> Iterable['Result']:
        for result in self.result_list[entity][:self.TOPN]:
            yield result

    @property
    def precision(self) -> Precision:
        if self.number_of_found_unannotated_results > 0:
            message = '{} results have not been annotated; treating them as non-relevant'.format(
                self.number_of_found_unannotated_results)
            LOGGER.warning(message)
        if self.number_of_found_results == 0:
            if self.number_of_total_relevant_results > 0:
                return 0.0
            else:
                raise ValueError('There are no results in the result list and no known relevant results')
        precision = 1.0 * self.number_of_found_relevant_results / self.number_of_found_results
        return precision

    @property
    def recall(self) -> Recall:
        if self.number_of_found_unannotated_results > 0:
            message = '{} results have not been annotated; treating them as non-relevant'.format(
                self.number_of_found_unannotated_results)
            LOGGER.warning(message)
        if self.number_of_total_relevant_results == 0:
            return 1.0
        recall = 1.0 * self.number_of_found_relevant_results / self.number_of_total_relevant_results
        return recall

    def __str__(self) -> str:
        summary = 'The results achieved Precision@{topn} of {:.2f}%, Recall@{topn} of {:.2f}%, and F@{topn}-score of {:.2f}%.'
        summary = summary.format(100.0 * self.precision, 100.0 * self.recall, 100.0 * self.f_score, topn=self.TOPN)
        return summary

    def __repr__(self) -> str:
        return str(self)
